function [digits, nPoints, idxOpts] = parsenacainputs(optList, varargin)
    % PARSENACAINPUTS Parses the inputs for naca* functions and checks their validity.
    %   Returns the naca denomination, the number of points and the index where the remaining
    %   options start.
    %   If some inputs were not provided, the functions returns an empty array for those.
    %
    % -----
    % Usage:
    %   [digits, nPoints, idxOpts] = PARSENACAINPUTS (optList)
    %   [...] = PARSEPOLARINPUTS (optList, DIGITS)
    %   [...] = PARSEPOLARINPUTS (optList, DIGITS, NPOINTS)
    %   [...] = PARSEPOLARINPUTS (optList, DIGITS, NPOINTS, ...)
    %
    % -----
    % (c) Copyright 2022 University of Liege
    % Author: Thomas Lambert <t.lambert@uliege.be>
    % ULiege - Aeroelasticity and Experimental Aerodynamics
    % Apache 2.0 License
    % https://gitlab.uliege.be/am-dept/matlab_airfoil_toolbox

    % ----------------------------------------------------------------------------------------------

    import af_tools.utils.*

    % Constants and defaults
    DEFAULT_NPOINTS = 100;
    MAX_PROMPTS = 3; % Number of re-promts before giving up

    % First ensures the option list is properly set
    validateattributes(optList, {'char', 'string', 'cell'}, {}, mfilename(), 'optList', 1);

    % Initialize outputs
    digits = [];
    nPoints = DEFAULT_NPOINTS;

    % Parse inputs
    hasNoInput = isempty(varargin) || isoption(optList, varargin{1});
    if hasNoInput
        idxOpts = 1;
        count = 1;
        while isempty(digits) && count <= MAX_PROMPTS
            prompt = 'Input NACA 4 or 5 digits (ex: 0012, 24012): ';
            digits = input(prompt, 's');
            count = count + 1;
        end
        if count >= MAX_PROMPTS
            error('MATLAB:parsenacainputs:noUserInput', ...
                  'No data input after 3 attempts. Operation aborded.\n');
        end
    else
        idxOpts = 2;
        digits = varargin{1};

        if length(varargin) >= 2 && ~isoption(optList, varargin{2})
            idxOpts = 3;
            nPoints = varargin{2};
        end

    end

    % Validate inputs
    validatedigits(digits);
    validatenpoints(nPoints);

end

% --------------------------------------------------------------------------------------------------
function bool = isoption(optList, var)
    % ISOPTION Returns true if var is contained in option list

    bool = any(strcmpi(var, optList));
end

% --------------------------------------------------------------------------------------------------
function validatedigits(digits)
    % VALIDATEDIGITS Validates digit input.

    % Correct type
    validateattributes(digits, {'char', 'string'}, {'scalartext'}, mfilename(), 'digits', 1);
    if isstring(digits)
        digits = char(digits);
    end

    % Only accept NACA 4- or 5-digit
    if length(digits) < 4 || length(digits) > 5
        error('MATLAB:parsenacainputs:wrongNumberOfDigits', ...
              'nacacamber and nacaairfoil can only be used for NACA 4 or 5 digits airfoils.');
    end

    % Ensures the character vector only contains digits
    if ~all(isstrprop(digits, 'digit'))
        error('MATLAB:parsenacainputs:invalidDigits', ...
              'The input (''%s'') does not correspond to a proper NACA airfoil.', digits);
    end

    % Ensure the third number of a NACA-5 is either 0 or 1
    if length(digits) == 5
        q = str2double(digits(3));
        if q ~= 0 && q ~= 1
            error('MATLAB:parsenacainputs:invalid5Digits', ...
                  ['The third digit of a NACA 5 series must either be a 0 '...
                   '(standard camber) or a 1 (reflex camber). Found %d'], q);
        end
    end
end

% --------------------------------------------------------------------------------------------------
function validatenpoints(nPoints)
    % VALIDATENPOINTS Validates nPoints input.

    % Correct type
    validateattributes(nPoints, {'numeric'}, {'scalar', '>=', 2}, mfilename(), 'nPoints', 2);
    % Enough points to have a good result
    if nPoints <= 10
        warning('MATLAB:parsenacainputs:lowNPoints', ...
                ['The number of points is very low (%d), ' ...
                 'the camberline may be poorly defined.'], nPoints);
    end

end
