function [alphaRange, c_lRange, c_lSlope] = findcllinearrange(varargin)
    % FINDCLLINEARRANGE Finds the linear range of the lift coefficient its slope.
    %   This function approximates the lift coefficient linear range by looking at its first and
    %   second derivatives (using hard-coded tolerance values).
    %
    %   FINDCLLINEARRANGE accepts inputs in the form of a Polar structure (generated with XF2MAT) or
    %   values for angles of attack and their associated cl.
    %
    %   If the input is specified as aoa and cl, the two arrays must have the same number of rows
    %   (one for each angle of attack). The columns indicate values for different polars
    %   (different Mach, different Reynolds,...).
    %   For example, if the cl values are given for 2 different Reynols and there are 100 values for
    %   the aoa, cl must be [100 x 2].
    %
    %   If extended polars are used as input (see EXTENDPOLAR), the function will only look for the
    %   zero-lift angle in the range of [-30, 30] deg.
    %
    % <a href="https://gitlab.uliege.be/am-dept/matlab_airfoil_toolbox">Documentation (README)</a>
    % <a href="https://gitlab.uliege.be/am-dept/matlab_airfoil_toolbox/-/issues">Report an issue</a>
    % -----
    %
    % Usage:
    %   [ALPHA_RANGE, CL_RANGE, CL_SLOPE] = FINDCLLINEARRANGE prompts the user to input a Polar
    %   structure and determines the range of angles of attach (ALPHA_RANGE) in which the cl is
    %   linear. It also outputs the values of the lift coefficient in that range (CL_RANGE) and the
    %   slope of the linear cl (CL_SLOPE).
    %
    %   [ALPHA_RANGE, CL_RANGE, CL_SLOPE] = FINDCLLINEARRANGE(POLAR) uses the POLAR strcture given
    %   as input instead of prompting the user to select the file.
    %
    %   [ALPHA_RANGE, CL_RANGE, CL_SLOPE] = FINDCLLINEARRANGE(ALPHA, CL) uses directly the values
    %   given in ALPHA and CL to determine the linear range characteristics.
    %
    % Inputs:
    %   Polar  : Polar structure (e.g. output from XF2MAT)
    %   alpha  : Vector or array with angles of attack [aoas x 1] or [aoas x nbPolar]
    %   cl     : Array with lift coefficients [aoas x nbPolar]
    %
    % Output:
    %   alphaRange : Range of angles of attack where cl is considered linear, IN RADIANS
    %   clRange    : Range of lift coefficients where cl is considered linear
    %   clSlope    : Lift curve slope in the linear range
    %
    % Example:
    %   FINDCLLINEARRANGE
    %   FINDCLLINEARRANGE(Polar)
    %   FINDCLLINEARRANGE(alpha, cl)
    %
    % See also: XF2MAT, FINDSTALL, FINDZEROLIFT.
    %
    % -----
    % (c) Copyright 2022 University of Liege
    % Author: Thomas Lambert <t.lambert@uliege.be>
    % ULiege - Aeroelasticity and Experimental Aerodynamics
    % Apache 2.0 License
    % https://gitlab.uliege.be/am-dept/matlab_airfoil_toolbox

    % ----------------------------------------------------------------------------------------------

    narginchk(0, 2);

    % Import other functions from this package
    import af_tools.utils.*

    % Defaults and constants
    ORIGINAL_RANGE = deg2rad([-30; 30]);  % Look for values in this range only
    OPTION_LIST = {};   % No options for this function

    % Extract and validate the inputs
    [alpha, c_l] = parsepolarinputs(OPTION_LIST, varargin{:});

    % Trim extended polars to original range
    % (avoid issues if extended polars are used as input)
    [alpha, c_l] = trimtorange(ORIGINAL_RANGE, alpha, c_l);

    % -------------------------------------------
    % Find linear range for each polar

    [alphaRange, c_lRange, c_lSlope] = linearrangefinder(alpha, c_l);

end

% --------------------------------------------------------------------------------------------------
function [alphaRange, c_lRange, c_lSlope] = linearrangefinder(alpha, c_l)
    % FINDLINEARRANGEVECT Finds the linearrange and returns the corresponding range and slope.
    %   We consider that the linear range meets the two following criterions:
    %       - First derivative close to 2*pi (TOL_DCL defines how close)
    %       - Second derivative close to 0 (TOL_DDCL defines how close)

    % Constants and defaults
    IDEAL_LIFT_SLOPE = 2 * pi;  % Ideal Cl curve, [rad]
    TOL_DCL = 0.6;   % Tolerance on the slope
    TOL_DDCL = 0.01;  % Tolerance on the curvature

    % Initialization
    alphaRange = zeros(2, size(c_l, 2));
    c_lRange = zeros(2, size(c_l, 2));
    c_lSlope = zeros(1, size(c_l, 2));

    % Loop for every polar data (every column of cl)
    for i = 1:size(c_l, 2)
        [tmpRange, rangeFound] = findrange(i, alpha, c_l, TOL_DCL, TOL_DDCL, IDEAL_LIFT_SLOPE);

        % Check issues in case linear range was not found
        if ~rangeFound
            ploterror(alpha, c_l, i, tmpRange);
            error('MATLAB:findcllinearrange:noRangeFound', ...
                  ['findcllinearrange was not able to determine the linear range of the given '...
                   'polar.\nMaybe the polar does not have enough points, or it is malformed.\n '...
                   'See the attached plot for more details']);

        end

        c_lRange(:, i) = c_l(tmpRange, i);
        alphaRange(:, i) = alpha(tmpRange, i);
        c_lSlope(i) = diff(c_lRange(:, i)) / diff(alphaRange(:, i));

    end

end

% --------------------------------------------------------------------------------------------------
function bool = isbound(dcl, ddcl, dClTol, ddClTol, idealCl)
    % ISBOUND Returns true if the point is a valid bound for the linear range
    isSlopeOk = isslopeabovecla(dcl, dClTol, idealCl);
    isCurvOk = iscurvaturezero(ddcl, ddClTol);

    bool = isSlopeOk && isCurvOk;
end

% --------------------------------------------------------------------------------------------------
function bool = isslopeabovecla(dCl, tol, idealCl)
    % ISSLOPEABOVECLA Returns true if the slope is larger than the ideal slope
    bool = dCl > tol * idealCl;
end

% --------------------------------------------------------------------------------------------------
function bool = iscurvaturezero(ddCl, tol)
    % ISCURVATUREZERO Returns true if the curvature is almost zero
    bool = abs(ddCl) < tol;
end

% --------------------------------------------------------------------------------------------------
function ploterror(alpha, c_l, i, tmpRange)
    % PLOTERROR Plots the polar in case of error

    figure('Name', 'linearrange:Debug');
    hold on;
    plot(alpha(:, i), c_l(:, i), '--');
    plot(alpha(tmpRange(1):tmpRange(2), i), c_l(tmpRange(1):tmpRange(2), i), ...
         'color', 'r', 'lineWidth', 1.2);
    plot(alpha(tmpRange(1), i), c_l(tmpRange(1), i), ...
         's', 'MarkerFaceColor', 'b', 'Markersize', 12);
    plot(alpha(tmpRange(2), i), c_l(tmpRange(2), i), ...
         's', 'MarkerFaceColor', 'b', 'Markersize', 12);
    grid on;
    legend('Polar', 'Attempt at linear range', 'Missing point(s)');

end

% --------------------------------------------------------------------------------------------------
function [tmpRange, rangeFound] = findrange(i, alpha, c_l, TOL_DCL, TOL_DDCL, IDEAL_LIFT_SLOPE)
    % FINDRANGE Finds the range iteratively.

    rangeFound = false;

    % First and second derivatives
    dalpha = diff(alpha(:, i));
    dcl = diff(c_l(:, i)) ./ dalpha(1);
    ddcl = diff(dcl) ./ dalpha(1);

    % Adapt tolerance to angle of attack
    ddClTol = TOL_DDCL / dalpha(1)^2;

    % Initialize linear range start and end (mostly useful for debugging)
    tmpRange = [1, length(alpha(:, i))];

    maxFound = false;
    minFound = false;

    for j = 1:(length(alpha)) - 2

        % Check if current point is a valid point of the linear range
        isMin = isbound(dcl(j), ddcl(j), TOL_DCL, ddClTol, IDEAL_LIFT_SLOPE);
        isMax = isbound(dcl(end + 1 - j), ddcl(end + 1 - j), TOL_DCL, ddClTol, ...
                        IDEAL_LIFT_SLOPE);

        if isMin && ~minFound
            minFound = true;
            tmpRange(1) = j;
        end

        if isMax && ~maxFound
            maxFound = true;
            tmpRange(end) = length(dcl) + 2 - j;
        end

        if minFound && maxFound
            rangeFound = true;
            break
        end

    end

end
