function Polar = xf2mat(varargin)
    % XF2MAT Aggregates multiple XFOIL or XFLR5 polar results into a single structure.
    %   This function aggregates different airfoil polars obtained with XFOIL or XFLR5 (saved as
    %   textfiles) into a single structure for easier handling in other scripts.
    %
    %   This function can be run without arguments, in this case it will prompt the user for the
    %   input. Otherwhile, the function asks for two main arguments:
    %       1. inputDir: the directory where to look for the files
    %       2. inputFiles: the files to load
    %
    %   The function also comes with options to save automatically the resulting Polar structure and
    %   to trim all results to the same range of angles of attack.
    %
    % Note:
    %   If multiple textfiles are selected as input, by default, the function will extend every
    %   polar to the maximum range of angles of attack found. Polars with fewer data will be
    %   assigned NaN values at the missing angles of attack.
    %   If the 'trimAoas' is set to true, the function will find the largest common range of angles
    %   of attack accross the various input polars and trim everything to that range. This prevents
    %   NaN values in the polars, but removes some data from the most complete inputs.
    %
    % <a href="https://gitlab.uliege.be/thlamb/airfoil_toolbox">Documentation (README)</a>
    % <a href="https://gitlab.uliege.be/thlamb/airfoil_toolbox/-/issues">Report an issue</a>
    % -----
    %
    % Usage:
    %   POLAR = XF2MAT prompts the user for all inputs, then aggregate and convert the polar in a
    %   single POLAR structure.
    %
    %   POLAR = XF2MAT(INPUTDIR) aggregates all polars found in INPUTDIR into a single Polar
    %   structure.
    %
    %   POLAR = XF2MAT(INPUTDIR, INPUTFILES) aggregates only the polars specified by INPUTFILES
    %   found in INPUTDIR into a single Polar structure.
    %
    %   POLAR = XF2MAT(..., 'autosave', true) saves the resulting POLAR structure automatically in
    %   a MAT-file.
    %
    %   POLAR = XF2MAT(..., 'trimAoas', true) trims all input polars to the largest common range of
    %   angles of attack to prevent NaN values in the ouput.
    %
    % Inputs:
    %   inputDir   : Path of the directory with the XFOIL/XFLR5 data
    %   inputFiles : Files to select (ex: '*' (default), '*0012*', {'file1','file2'}, etc)
    %
    % Output:
    %   Polar : Structure collecting the results
    %           - Polar.alpha    : Angles of attack [Rad]
    %           - Polar.airfoil  : Airfoil name
    %           - Polar.reynolds : Reynolds number [-]
    %           - Polar.mach     : Mach number [-]
    %           - Polar.nCrit    : BL trnansition factor [-]
    %           - Polar.cl       : Lift coefficient [-]
    %           - Polar.cd       : Drag coefficient [-]
    %           - Polar.cm       : Moment coefficient [-]
    %           - Polar.origin   : XFOIL / XFLR5
    %
    % Example:
    %   XF2MAT
    %   XF2MAT('test_data')
    %   XF2MAT('test_data', '*0012*')
    %   XF2MAT('test_data', {'0012_re1e5', '0012_re1e6'})
    %   XF2MAT('test_data', 'autosave', true)
    %   XF2MAT('test_data', '*', 'trimAoas', true)
    %
    % See also: EXTENDPOLAR, PLOTPOLARS.
    %
    % -----
    % XFOIL: http://web.mit.edu/drela/Public/web/xfoil/
    % XFLR5: https://www.xflr5.tech/xflr5.htm
    % -----
    % (c) Copyright 2022 University of Liege
    % Author: Thomas Lambert <t.lambert@uliege.be>
    % ULiege - Aeroelasticity and Experimental Aerodynamics
    % Apache 2.0 License
    % https://gitlab.uliege.be/am-dept/matlab_airfoil_toolbox

    % ----------------------------------------------------------------------------------------------

    narginchk(0, 6);

    % Import other functions from this package
    import af_tools.utils.*

    % Constants and defaults
    OPTION_LIST = {'autosave', 'trimAoas'};
    FILETYPE = '.txt';

    % Parse inputs
    [allFileNames, fullpath, idxOpts] = parsefileinputs(OPTION_LIST, FILETYPE, varargin{:});
    [autosave, trimAoas] = parseoptioninputs(varargin{idxOpts:end});

    % Convert filenames to string array
    allFileNames = string(allFileNames);

    % -------------------------------------------
    % Process XFoil/XFLR5 results

    % Number of files found
    nbFiles = length(allFileNames);

    % Initialize structure
    Tmp(nbFiles).resultArray = [];

    % Initialize AOA range
    if trimAoas
        aoaRange = [-180, 180];
    else
        aoaRange = [-3, 3];
    end

    for i = 1:nbFiles

        % Load file and extract parameters and results
        fileID = fopen(fullfile(fullpath, allFileNames{i}));

        % Scan file header to determine if it is from XFOIL or XFLR5
        % (this also position the pointer at the correct location for reading the rest)
        progName = [];
        while isempty(progName)
            tmpProg = textscan(fileID, '%s', 1, 'Delimiter', '\n');
            tmpProgname = char(tmpProg{1});
            if ~isempty(tmpProgname) && strcmp(tmpProgname(1:5), 'XFOIL')
                progName = 'XFOIL';
            elseif ~isempty(tmpProgname) && strcmp(tmpProgname(1:5), 'xflr5')
                progName = 'XFLR5';
            end
        end
        Polar.origin{1, i} = progName;

        % Parse the rest of the results
        % (! textscan works as a pointer, headerlines are from the previous line, not the top)
        tmpAirfoil = textscan(fileID, '%s %s', 1, 'Delimiter', '\n:', 'HeaderLines', 1);
        tmpParam = textscan(fileID, '%s %s %f %s %s %f %s %f %s %s %f %*[^\n]', ...
                            1, 'HeaderLines', 4);
        Tmp(i).resultArray = cell2mat(textscan(fileID, '%f %f %f %f %f %*[^\n]', ...
                                               'HeaderLines', 4));
        fclose(fileID);
        clear fileID;

        % Extract the airfoil name
        airfoil_name = char(tmpAirfoil{2});
        Polar.airfoil{1, i} = sanitizestring(airfoil_name);

        % Get Mach, Reynolds and Ncrit numbers corresponding to the processed file
        Polar.mach(1, i) = tmpParam{3};
        Polar.reynolds(1, i) = tmpParam{6} * 10^tmpParam{8};
        Polar.nCrit(1, i) = tmpParam{11};

        % Get the larger common AOA range and the smallest step
        if trimAoas
            aoaRange = findcommonrange(aoaRange, Tmp(i).resultArray);
        else
            aoaRange = findmaxrange(aoaRange, Tmp(i).resultArray);
        end

    end

    % Interpolate results over proper range of AOA if needed
    [Polar.alpha, Polar.cl, Polar.cd, Polar.cm] = standardizepolar(Tmp, aoaRange);
    clear Tmp;

    % Sort everything by increasing Reynolds
    [Polar.reynolds, indSort] = sort(Polar.reynolds);
    Polar.origin(:) = Polar.origin(indSort);
    Polar.airfoil(:) = Polar.airfoil(indSort);
    Polar.mach(:)  = Polar.mach(indSort);
    Polar.nCrit(:) = Polar.nCrit(indSort);
    Polar.cl(:, :)  = Polar.cl(:, indSort);
    Polar.cd(:, :)  = Polar.cd(:, indSort);
    Polar.cm(:, :)  = Polar.cm(:, indSort);

    % -------------------------------------------
    % Save results automatically
    if autosave
        filename = genfilename(Polar);
        save(filename, 'Polar');
    end

end

% --------------------------------------------------------------------------------------------------
function [autosave, trimAoas] = parseoptioninputs(varargin)
    % PARSEOPTIONINPUTS Parses the options and checks their validity

    import af_tools.utils.*

    % Constants and defaults
    DEFAULT_AUTOSAVE = false;    % Do not autosave by default
    DEFAULT_TRIMAOAS = false;    % Do not trim AOAs by default

    % Option validator
    validLogical = @(x) validateattributes(x, {'logical'}, {'scalar'}, mfilename());

    % Parse options
    p = inputParser;
    addParameter(p, 'autosave', DEFAULT_AUTOSAVE, validLogical);
    addParameter(p, 'trimAoas', DEFAULT_TRIMAOAS, validLogical);
    parse(p, varargin{:});
    autosave = p.Results.autosave;
    trimAoas = p.Results.trimAoas;

end

% --------------------------------------------------------------------------------------------------
function aoaRange = findcommonrange(aoaRange, resArray)
    % FINDCOMMONRANGE Find the best common range accors all polars

    aoa_step = min(diff(aoaRange));

    if resArray(1, 1) > aoaRange(1)
        aoaRange(1) = resArray(1, 1);
    end
    if resArray(end, 1) < aoaRange(end)
        aoaRange(end) = resArray(end, 1);
    end
    if min(diff(resArray(:, 1))) < aoa_step
        aoa_step = min(diff(resArray(:, 1)));
    end

    aoaRange = aoaRange(1):aoa_step:aoaRange(end);

end

% --------------------------------------------------------------------------------------------------
function aoaRange = findmaxrange(aoaRange, resArray)
    % FINDMAXRANGE Find the max range between all polars

    aoa_step = min(diff(aoaRange));

    if resArray(1, 1) < aoaRange(1)
        aoaRange(1) = resArray(1, 1);
    end
    if resArray(end, 1) > aoaRange(end)
        aoaRange(end) = resArray(end, 1);
    end
    if min(diff(resArray(:, 1))) < aoa_step
        aoa_step = min(diff(resArray(:, 1)));
    end

    aoaRange = aoaRange(1):aoa_step:aoaRange(end);

end

% --------------------------------------------------------------------------------------------------
function [alpha, c_l, c_d, c_m] = standardizepolar(Input, aoaRange)
    % STANDARDIZEPOLAR Standardize polar output.
    %   Re-interpolate everything over the same range of AOAs

    alpha = zeros(length(aoaRange), length(Input));
    c_l = zeros(length(aoaRange), length(Input));
    c_d = zeros(length(aoaRange), length(Input));
    c_m = zeros(length(aoaRange), length(Input));

    for i = 1:length(Input)
        alpha(:, i) = deg2rad(aoaRange');
        c_l(:, i) = interp1(Input(i).resultArray(:, 1), Input(i).resultArray(:, 2), aoaRange);
        c_d(:, i) = interp1(Input(i).resultArray(:, 1), Input(i).resultArray(:, 3), aoaRange);
        c_m(:, i) = interp1(Input(i).resultArray(:, 1), Input(i).resultArray(:, 5), aoaRange);
    end
end

% --------------------------------------------------------------------------------------------------
function filename = genfilename(Polar)

    % GENFILENAME Generates a proper filename for the saved file

    import af_tools.utils.appendextension

    order = floor(log10(Polar.reynolds));
    reBase = round(Polar.reynolds ./ 10.^order);

    if length(Polar.reynolds) == 1
        reStr = sprintf('-Re_%de%d', reBase, order);
    else
        reStr = sprintf('-Re_%de%d-%de%d', reBase(1), order(1), reBase(end), order(end));
    end
    filename_base = [Polar.airfoil{1}, reStr];

    % Prevent overwriting files
    filename = filename_base;
    i = 0;
    while isfile(appendextension(filename, '.mat'))
        i = i + 1;
        filename = [filename_base, '_v', num2str(i)];
    end

    filename = appendextension(filename, '.mat');

end
